
module Ensemble(Ensemble,ensembleVide,(€),inserer,tailleReelle,taille,insererIntervalle,insererListe,supprimer,supprimerM,union,intersection,foldSet) where



-- Arbre Binaire de Recherche par Intervalles
data Ensemble a =  A (Ensemble a) a a (Ensemble a) | F
            deriving (Eq,Ord,Show,Read)

-- Creer un arbre vide
ensembleVide :: Ensemble a
ensembleVide = F

-- Rechercher une clef dans les intervalles de l'arbre
(€) :: (Enum a, Ord a) =>
       a -> Ensemble a -> Bool
_ € F = False
x € (A g l u d) | x < l =  x € g
                         | x > u = x € d
			 | otherwise = True

-- Inseerer une clef en modifiant potentielement les intervalles
inserer :: (Enum a, Ord a) =>
           Ensemble a -> a -> Ensemble a
inserer a x = insererIntervalle a x x

-- Somme des longueurs de tous les intervalles
tailleReelle :: Enum a =>
                Ensemble a -> Int
tailleReelle F = 0
tailleReelle (A g l u d) = tailleReelle  g + tailleReelle  d + fromEnum u - fromEnum l +1


-- Nombre de noeuds
taille :: Ensemble a -> Int
taille F = 0
taille (A g _ _ d) = taille  g + taille  d + 1

-- Inseerer tout un interval de clefs
insererIntervalle :: (Enum a, Ord a) =>
                     Ensemble a -> a -> a -> Ensemble a
insererIntervalle F x y = A F x y F
insererIntervalle (A g l u d) x y
   | y < (pred l) = A (insererIntervalle g x y) l u d
   | x > succ u = A g l u (insererIntervalle d x y)
   | x >= l && y <= u = A g l u d
   | x<l && y>u = let (g',x') = descenteDroite g x in
                  let (d',y') = descenteGauche d y in
                  A g' x' y' d'
   | y > u      = let (d',y') = descenteGauche d y in
                  A g l y' d'
   | x < l      = let (g',x') = descenteDroite g x in
                  A g' x' u d


descenteDroite F x = (F,x)
descenteDroite (A g l u d) x | x < l     = descenteDroite g x
                              | x > succ u   = let (d',x') = descenteDroite d x in (A g l u d',x')
                              | otherwise = (g,l)
descenteGauche F x = (F,x)
descenteGauche (A g l u d) x | x > u     = descenteGauche d x
                              | x < pred l   = let (g',x') = descenteGauche g x in (A g' l u d,x')
                              | otherwise = (d,l)


-- Insertion d'une liste de clefs
insererListe :: (Enum a, Ord a) =>
                Ensemble a -> [a] -> Ensemble a
insererListe = foldl inserer

-- Suppression d'une clef
supprimer :: (Enum a, Ord a) =>
             Ensemble a -> a -> Ensemble a
supprimer F _ = F
supprimer (A g l u d) x | x < l  = A (supprimer g x) l u d
                        | x > u  = A g l u (supprimer d x)
                        | x == l && x==u = fusionner  g d
                        | x == l =  A g (succ l) u d
                        | x == u =  A g l (pred u) d
                        | otherwise = A (A g l (pred x) F) (succ x) u d



fusionner g F = g
fusionner g (A gg l u d) = A (fusionner g gg) l u d


-- Suppression d'une clef avec resultat Nothing si clef non presente
supprimerM :: (Enum a, Ord a) =>
              Ensemble a -> a -> Maybe (Ensemble a)
supprimerM F _ = Nothing
supprimerM (A g l u d) x | x < l  = fmap (\g'->A g' l u d) (supprimerM g x)
                         | x > u  = fmap (\d'->A g l u d') (supprimerM d x)
                         | x == l && x==u = Just (supprimer_noeud  g d)
                         | x == l =  Just (A g (succ l) u d)
                         | x == u =  Just (A g l (pred u) d)
                         | otherwise = Just (A (A g l (pred x) F) (succ x) u d)


supprimer_noeud g F = g
supprimer_noeud g (A gg l u d) = A (supprimer_noeud g gg) l u d


-- union des deux ensembles
union :: (Enum a, Ord a) =>
         Ensemble a -> Ensemble a -> Ensemble a
union = foldSet2 (\(l,u) e3 -> insererIntervalle e3 l u)


-- intersection des deux ensembles
intersection :: (Enum a, Ord a) =>
                Ensemble a -> Ensemble a -> Ensemble a
intersection F _ = F 
intersection _ F = F 
intersection e1@(A g l u d) e2@(A g' l' u' d')
   | l'<l && u'>u   =  A (intersection g e2) l u (intersection d e2)
   | l'<l && u'>=l  =  A (intersection g e2) l u' (intersection e1 d')
   | u'>u && l'<=u  =  A (intersection e1 g') l' u (intersection d e2)
   | l'>=l && u'<=u =  A (intersection e1 g') l' u' (intersection e1 d')
   | l'>u           =  case d  of     
                           F -> intersection e1  g'
	                   (A dg dl du dd) -> unionDisjointeG (intersection (A g l u F)  g') (unionDisjointeD (intersection d g') (intersection d (A F l' u' d')))
   | otherwise      =  case d' of
                          F -> intersection g e2
	                  (A dg' dl' du' dd') -> union (intersection g e2) (intersection e1 d')  


-- union en supposant les elements de e1 plus petit que ceux de e2 plus petit que ceux de e3
unionDisjointeD F e2 = e2 
unionDisjointeD (A g l u d) e2 = A g l u (unionDisjointeD d e2)
unionDisjointeG e1 F = e1 
unionDisjointeG e1 (A g l u d) = A (unionDisjointeG e1 g) l u d



-- parcourt la structure en profondeur comme le foldl parcourt une liste
foldSet :: (Enum a) => (b -> a -> b) -> b -> Ensemble a -> b 
foldSet f x F = x
foldSet f x (A g l u d) = foldSet f (foldl f (foldSet f x g) [l..u]) d


-- parcourt la structure en profondeur comme le foldl parcourt une liste mais sur des intervalles
foldSet2 :: (Enum a, Ord a) => ((a,a) -> b -> b) -> b -> Ensemble a -> b 
foldSet2 f x F = x
foldSet2 f x (A g l u d) = foldSet2 f (foldSet2 f (f (l,u) x) g) d